;; Generated package description from minimail.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "minimail" "0.4.2.0.20260511.125743" "Simple, non-blocking IMAP email client" '((emacs "30.1") (compat "31") (transient "0.12")) :commit "b36a819cef7e17c9cf6a89130495e41a71b3a3ad" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("mail") :url "https://codeberg.org/astoff/minimail")
